/*
 * Copyright (c) 2010 Samsung Electronics Co., Ltd.
 *              http://www.samsung.com/
 *
 * S5PC110 - LCD Driver for U-Boot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <common.h>
#include <asm/io.h>
#include <asm/types.h>
#include <lcd.h>
#include <video_font.h>

DECLARE_GLOBAL_DATA_PTR;

#define LCD_BGCOLOR		0x1428A0
#define LCD_BPP                           LCD_COLOR8


static unsigned int gFgColor = 0xFF;
static unsigned int gLeftColor = LCD_BGCOLOR;

#define Inp32(_addr)		readl(_addr)
#define Outp32(addr, data)	(*(volatile u32 *)(addr) = (data))
#define Delay(_a)		udelay(_a*1000)


#if defined(CFG_LCD_LB070WV6)
/*
 * LB070WV6
 */

#define LCD_WIDTH		800
#define LCD_HEIGHT		480
#define LCD_VBPD		14
#define LCD_VFPD		16
#define LCD_VSPW		1
#define LCD_HBPD		64
#define LCD_HFPD		64
#define LCD_HSPW		128

vidinfo_t panel_info = {
        .vl_col = 800,
        .vl_row = 480,
        .vl_bpix = 3,
};

int lcd_line_length;
int lcd_color_fg;
int lcd_color_bg;

void *lcd_base;                         /* Start of framebuffer memory  */
void *lcd_console_address;              /* Start of console buffer      */

short console_col = 5;
short console_row = 5;

void LCD_Initialize_LB070WV6(void)
{
	u32 uFbAddr = CFG_LCD_FBUFFER;

	u32 i,uTmpReg;
	u32* pBuffer = (u32*)uFbAddr;


//	LCD_setleftcolor(0x0);
//	LCD_setfgcolor(0x0);
	LCD_setprogress(0);


	uTmpReg = Inp32(0xE0107008);		// need to be changed later (09.01.23 WJ.Kim)	
	uTmpReg = (uTmpReg & ~(0x3)) | 0x2 ;
	Outp32(0xE0107008, uTmpReg);

	uTmpReg = Inp32(0xf800027c);
	uTmpReg = (uTmpReg & ~(0x3)) | 0 ;
	Outp32(0xf800027c,uTmpReg);


	Outp32(0xE0200120, 0x22222222);	//set GPF0 as LVD_HSYNC,VSYNC,VCLK,VDEN,VD[3:0]
	Outp32(0xE0200128,0x0);			//set pull-up,down disable
	Outp32(0xE0200140, 0x22222222);	//set GPF1 as VD[11:4]
	Outp32(0xE0200148,0x0);			//set pull-up,down disable
	Outp32(0xE0200160, 0x22222222);	//set GPF2 as VD[19:12]
	Outp32(0xE0200168,0x0);			//set pull-up,down disable
	Outp32(0xE0200180, 0x00002222);	//set GPF3 as VD[23:20]
	Outp32(0xE0200188,0x0);			//set pull-up,down disable

	Outp32(0xe02000A0, 0x11111111);	//set GPD0 as Backlight
	Outp32(0xe02000A8,0x0);			//set pull-up,down disable
	Outp32(0xe02000A4,(1<<0));

	Outp32(0xE0200080, 0x11111111);	//set GPC1 as LCD_PWREN
	Outp32(0xE0200088,0x0);			//set pull-up,down disable
	Outp32(0xE0200084,(1<<4));

	Outp32(0xf8000004, 0x60);
	Outp32(0xf8000010, ((LCD_VBPD-1)<<16)|((LCD_VFPD-1)<<8)|(LCD_VSPW-1));
	Outp32(0xf8000014, ((LCD_HBPD-1)<<16)|((LCD_HFPD-1)<<8)|(LCD_HSPW-1));
	Outp32(0xf8000170, 0x0);
	Outp32(0xf8000018, ((LCD_HEIGHT-1)<<11)|(LCD_WIDTH-1));
	Outp32(0xf8000000, 0x0);
	Outp32(0xf8000000, 0x190);
	//Outp32(0xf8000000, 0x110);
	Outp32(0xf8000130, 0x20);
	Outp32(0xf8000020, 0x0);
	Outp32(0xf8000024, 0x0);
	Outp32(0xf8000028, 0x0);
	Outp32(0xf800002c, 0x0);
	Outp32(0xf8000030, 0x0);
	Outp32(0xf8000034, 0x0);
	Outp32(0xf8000180, 0x0);
	Outp32(0xf8000184, 0x0);
	Outp32(0xf8000188, 0x0);
	Outp32(0xf800018c, 0x0);
	Outp32(0xf8000190, 0x0);
	Outp32(0xf8000140, 0x0);
	Outp32(0xf8000148, 0x0);
	Outp32(0xf8000150, 0x0);
	Outp32(0xf8000158, 0x0);
	Outp32(0xf8000058, 0x0);
	Outp32(0xf8000208, 0x0);
	Outp32(0xf800020c, 0x0);
	Outp32(0xf8000068, 0x0);
	Outp32(0xf8000210, 0x0);
	Outp32(0xf8000214, 0x0);
	Outp32(0xf8000078, 0x0);
	Outp32(0xf8000218, 0x0);
	Outp32(0xf800021c, 0x0);
	Outp32(0xf8000088, 0x0);
	Outp32(0xf8000220, 0x0);
	Outp32(0xf8000224, 0x0);
	Outp32(0xf8000260, 0x1);
	Outp32(0xf8000034, 0x0);
	Outp32(0xf80000a4, uFbAddr + 0x0);
	Outp32(0xf80000d4, uFbAddr + (LCD_WIDTH*LCD_HEIGHT*4));
	Outp32(0xf80000a0, uFbAddr + 0x0);
	Outp32(0xf80000d0, uFbAddr + (LCD_WIDTH*LCD_HEIGHT*4));
	Outp32(0xf80020a0, uFbAddr + 0x0);
	Outp32(0xf80020d0, uFbAddr + (LCD_WIDTH*LCD_HEIGHT*4));
	Outp32(0xf8000100, (LCD_WIDTH*4));
	Outp32(0xf8000020, 0x802c);
	Outp32(0xf8000040, 0x0);
	Outp32(0xf8000044, ((LCD_WIDTH-1)<<11)|(LCD_HEIGHT-1));
	Outp32(0xf8000200, 0xffffff);
	Outp32(0xf8000204, 0xffffff);
	Outp32(0xf8000020, 0x802d);
	Outp32(0xf8000034, 0x1);

	Outp32(0xf8000000, 0x193);
	//Outp32(0xf8000000, 0x113);
}
#else
// "No LCD Type is defined!"
#endif

/*
 */

/************************************************************************/
/**  Small utility to check that you got the colours right              */
/************************************************************************/
#define N_BLK_VERT      2
#define N_BLK_HOR       3

static int test_colors[N_BLK_HOR*N_BLK_VERT] = {
        CONSOLE_COLOR_RED,      CONSOLE_COLOR_GREEN,    CONSOLE_COLOR_YELLOW,
        CONSOLE_COLOR_BLUE,     CONSOLE_COLOR_MAGENTA,  CONSOLE_COLOR_CYAN,
};

static void test_pattern (void)
{
        ushort v_max  = panel_info.vl_row;
        ushort h_max  = panel_info.vl_col;
        ushort v_step = (v_max + N_BLK_VERT - 1) / N_BLK_VERT;
        ushort h_step = (h_max + N_BLK_HOR  - 1) / N_BLK_HOR;
        ushort v, h;
        uchar *pix = (uchar *)lcd_base;

        printf ("[LCD] Test Pattern: %d x %d [%d x %d]\n",
                h_max, v_max, h_step, v_step);

        /* WARNING: Code silently assumes 8bit/pixel */
        for (v=0; v<v_max; ++v) {
                uchar iy = v / v_step;
                for (h=0; h<h_max; ++h) {
                        uchar ix = N_BLK_HOR * iy + (h/h_step);
                        *pix++ = test_colors[ix];
                }
        }
}

void lcd_ctrl_init (void *lcdbase)
{
	lcd_line_length = (panel_info.vl_col * NBITS (panel_info.vl_bpix)) / 8;
	lcd_base = (void *)(gd->fb_base);
	LCD_Initialize_LB070WV6();
}

void LCD_testpattern(void)
{
	test_pattern();
}
void LCD_turnon(void)
{
	lcd_line_length = (panel_info.vl_col * NBITS (panel_info.vl_bpix)) / 8;
	//printf("lcd_line_length = %d \n",lcd_line_length);
 	lcd_base = (u32*)CFG_LCD_FBUFFER;
	LCD_Initialize_LB070WV6();
}

void LCD_setfgcolor(unsigned int color)
{
	lcd_color_fg = color;
}

void LCD_setleftcolor(unsigned int color)
{
	lcd_color_bg = color;
}

void LCD_setprogress(int percentage)
{
	u32 i, j;
	u32* pBuffer = (u32*)CFG_LCD_FBUFFER;
#if 1
	for (i=0; i < (LCD_HEIGHT/100)*percentage; i++)
	{
		for (j=0; j < LCD_WIDTH; j++)
		{
			*pBuffer++ = lcd_color_fg;
		}
	}

	for (; i < LCD_HEIGHT; i++)
	{
		for (j=0; j < (LCD_WIDTH >> 5); j++)
		{
			*pBuffer++ = lcd_color_bg;
		}
		for (; j < LCD_WIDTH; j++)
		{
			*pBuffer++ = lcd_color_fg;
		}
	}
#endif
	for(i=0;i<LCD_HEIGHT;i++)
		for(j=0;j<LCD_WIDTH;j++)
			*pBuffer++ = gFgColor;
}


void lcd_setcolreg (ushort regno, ushort red, ushort green, ushort blue)
{
	printf("%s..\n",__func__);
 
}
void lcd_enable(void)
{
	printf("%s..\n",__func__);

}

void lcd_disable(void)
{
	printf("%s..\n",__func__);

}

ulong calc_fbsize (void)
{
    ulong size;
	printf("%s..\n",__func__);
 
    int line_length = (panel_info.vl_col * panel_info.vl_bpix) / 8;
    size = line_length * panel_info.vl_row;
 
    return size;
}

static void lcd_drawchars (ushort x, ushort y, uchar *str, int count)
{
        uchar *dest;
        ushort off, row;

	printf("%s..\n",__func__);
        dest = (uchar *)(lcd_base + y * lcd_line_length + x * (1 << LCD_BPP) / 8);
        off  = x * (1 << LCD_BPP) % 8;

        for (row=0;  row < VIDEO_FONT_HEIGHT;  ++row, dest += lcd_line_length)  {
                uchar *s = str;
                uchar *d = dest;
                int i;

#if LCD_BPP == LCD_MONOCHROME
                uchar rest = *d & -(1 << (8-off));
                uchar sym;
#endif
                for (i=0; i<count; ++i) {
                        uchar c, bits;

                        c = *s++;
                        bits = video_fontdata[c * VIDEO_FONT_HEIGHT + row];

#if LCD_BPP == LCD_MONOCHROME
                        sym  = (COLOR_MASK(lcd_color_fg) & bits) |
                               (COLOR_MASK(lcd_color_bg) & ~bits);

                        *d++ = rest | (sym >> off);
                        rest = sym << (8-off);
#elif LCD_BPP == LCD_COLOR8
			printf("c = %x d  = %x bits = %x \n", c, *dest, bits);
                        for (c=0; c<8; ++c) {
                                *d++ = (bits & 0x80) ?
                                                lcd_color_fg : lcd_color_bg;
                                bits <<= 1;
                        }
#elif LCD_BPP == LCD_COLOR16
                        for (c=0; c<16; ++c) {
                                *d++ = (bits & 0x80) ?
                                                lcd_color_fg : lcd_color_bg;
                                bits <<= 1;
                        }
#endif
                }
#if LCD_BPP == LCD_MONOCHROME
                *d  = rest | (*d & ((1 << (8-off)) - 1));
#endif
        }
}

static inline void lcd_putc_xy (ushort x, ushort y, uchar c)
{
	printf("%s..\n",__func__);
#if defined(CONFIG_LCD_LOGO) && !defined(CONFIG_LCD_INFO_BELOW_LOGO)
        lcd_drawchars (x, y+BMP_LOGO_HEIGHT, &c, 1);
#else
        lcd_drawchars (x, y, &c, 1);
#endif
}

static void console_scrollup (void)
{
	printf("%s..\n",__func__);
#if 1
        /* Copy up rows ignoring the first one */
        memcpy (CONSOLE_ROW_FIRST, CONSOLE_ROW_SECOND, CONSOLE_SCROLL_SIZE);

        /* Clear the last one */
        memset (CONSOLE_ROW_LAST, COLOR_MASK(lcd_color_bg), CONSOLE_ROW_SIZE);
#else
        /*
         * Poor attempt to optimize speed by moving "long"s.
         * But the code is ugly, and not a bit faster :-(
         */
        ulong *t = (ulong *)CONSOLE_ROW_FIRST;
        ulong *s = (ulong *)CONSOLE_ROW_SECOND;
        ulong    l = CONSOLE_SCROLL_SIZE / sizeof(ulong);
        uchar  c = lcd_color_bg & 0xFF;
        ulong val= (c<<24) | (c<<16) | (c<<8) | c;

        while (l--)
                *t++ = *s++;

        t = (ulong *)CONSOLE_ROW_LAST;
        l = CONSOLE_ROW_SIZE / sizeof(ulong);

        while (l-- > 0)
                *t++ = val;
#endif
}

static inline void console_newline (void)
{
        ++console_row;
        console_col = 0;
	printf("%s..\n",__func__);

        /* Check if we need to scroll the terminal */
        if (console_row >= CONSOLE_ROWS) {
                /* Scroll everything up */
                console_scrollup () ;
                --console_row;
        }
}

void lcd_putc (const char c)
{
	printf("%s..\n",__func__);

        switch (c) {
#if 0
        case '\r':      console_col = 0;
                        return;

        case '\n':      console_newline();
                        return;

        case '\t':      /* Tab (8 chars alignment) */
                        console_col |=  8;
                        console_col &= ~7;

                        if (console_col >= CONSOLE_COLS) {
                                console_newline();
                        }
                        return;

        case '\b':      console_back();
                        return;
#endif

        default:        lcd_putc_xy (console_col * VIDEO_FONT_WIDTH,
                                     console_row * VIDEO_FONT_HEIGHT,
                                     c);
                        if (++console_col >= CONSOLE_COLS) {
                                console_newline();
                        }
                        return;
        }
        /* NOTREACHED */
}
void lcd_puts (const char *s)
{
	printf("%s..\n",__func__);

        while (*s) {
                lcd_putc (*s++);
        }
}
